import asyncio
import logging
from abc import ABC
from collections.abc import AsyncIterator, Sequence
from typing import Any, ClassVar, Generic, cast, final
from uuid import uuid4

from pydantic import BaseModel, TypeAdapter
from tenacity import RetryCallState, retry, stop_after_attempt, wait_random_exponential

from .errors import InputValidationError
from .generics_utils import AutoInstanceAttributesMixin
from .memory import MemT
from .packet import Packet
from .run_context import CtxT, RunContext
from .typing.events import Event, PacketEvent, ProcOutputEvent
from .typing.io import InT, OutT_co, ProcName
from .typing.tool import BaseTool

logger = logging.getLogger(__name__)


def retry_error_callback(retry_state: RetryCallState) -> None:
    exception = retry_state.outcome.exception() if retry_state.outcome else None
    if exception:
        if retry_state.attempt_number == 1:
            logger.warning(f"\nParallel run failed:\n{exception}")
        if retry_state.attempt_number > 1:
            logger.warning(f"\nParallel run failed after retrying:\n{exception}")


def retry_before_sleep_callback(retry_state: RetryCallState) -> None:
    exception = retry_state.outcome.exception() if retry_state.outcome else None
    logger.info(
        f"\nRetrying parallel run (attempt {retry_state.attempt_number}):\n{exception}"
    )


class Processor(AutoInstanceAttributesMixin, ABC, Generic[InT, OutT_co, MemT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self, name: ProcName, num_par_run_retries: int = 0, **kwargs: Any
    ) -> None:
        self._in_type: type[InT]
        self._out_type: type[OutT_co]

        super().__init__()

        self._in_type_adapter: TypeAdapter[InT] = TypeAdapter(self._in_type)
        self._out_type_adapter: TypeAdapter[OutT_co] = TypeAdapter(self._out_type)

        self._name: ProcName = name
        self._memory: MemT
        self._num_par_run_retries: int = num_par_run_retries

    @property
    def in_type(self) -> type[InT]:
        return self._in_type

    @property
    def out_type(self) -> type[OutT_co]:
        return self._out_type

    @property
    def name(self) -> ProcName:
        return self._name

    @property
    def memory(self) -> MemT:
        return self._memory

    @property
    def num_par_run_retries(self) -> int:
        return self._num_par_run_retries

    def _validate_and_resolve_single_input(
        self,
        chat_inputs: Any | None = None,
        in_packet: Packet[InT] | None = None,
        in_args: InT | None = None,
    ) -> InT | None:
        multiple_inputs_err_message = (
            "Only one of chat_inputs, in_args, or in_message must be provided."
        )
        if chat_inputs is not None and in_args is not None:
            raise InputValidationError(multiple_inputs_err_message)
        if chat_inputs is not None and in_packet is not None:
            raise InputValidationError(multiple_inputs_err_message)
        if in_args is not None and in_packet is not None:
            raise InputValidationError(multiple_inputs_err_message)

        if in_packet is not None:
            if len(in_packet.payloads) != 1:
                raise InputValidationError(
                    "Single input runs require exactly one payload in in_packet."
                )
            return in_packet.payloads[0]
        return in_args

    def _validate_and_resolve_parallel_inputs(
        self,
        chat_inputs: Any | None,
        in_packet: Packet[InT] | None,
        in_args: Sequence[InT] | None,
    ) -> Sequence[InT]:
        if chat_inputs is not None:
            raise InputValidationError(
                "chat_inputs are not supported in parallel runs. "
                "Use in_packet or in_args."
            )
        if in_packet is not None:
            if not in_packet.payloads:
                raise InputValidationError(
                    "Parallel runs require at least one input payload in in_packet."
                )
            return in_packet.payloads
        if in_args is not None:
            return in_args
        raise InputValidationError(
            "Parallel runs require either in_packet or in_args to be provided."
        )

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | None = None,
        memory: MemT,
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> Sequence[OutT_co]:
        if in_args is None:
            raise InputValidationError(
                "Default implementation of _process requires in_args"
            )

        return cast("Sequence[OutT_co]", in_args)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | None = None,
        memory: MemT,
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        if in_args is None:
            raise InputValidationError(
                "Default implementation of _process requires in_args"
            )
        outputs = cast("Sequence[OutT_co]", in_args)
        for out in outputs:
            yield ProcOutputEvent(data=out, name=self.name)

    def _validate_outputs(self, out_payloads: Sequence[OutT_co]) -> Sequence[OutT_co]:
        return [
            self._out_type_adapter.validate_python(payload) for payload in out_payloads
        ]

    def _generate_run_id(self, run_id: str | None) -> str:
        if run_id is None:
            return str(uuid4())[:6] + "_" + self.name
        return run_id

    async def _run_single(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT_co]:
        resolved_in_args = self._validate_and_resolve_single_input(
            chat_inputs=chat_inputs, in_packet=in_packet, in_args=in_args
        )
        _memory = self.memory.model_copy(deep=True) if forgetful else self.memory
        outputs = await self._process(
            chat_inputs=chat_inputs,
            in_args=resolved_in_args,
            memory=_memory,
            run_id=self._generate_run_id(run_id),
            ctx=ctx,
        )
        val_outputs = self._validate_outputs(outputs)

        return Packet(payloads=val_outputs, sender=self.name)

    def _generate_par_run_id(self, run_id: str | None, idx: int) -> str:
        return f"{self._generate_run_id(run_id)}/{idx}"

    async def _run_par(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: Sequence[InT] | None = None,
        run_id: str | None = None,
        forgetful: bool = False,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT_co]:
        par_inputs = self._validate_and_resolve_parallel_inputs(
            chat_inputs=chat_inputs, in_packet=in_packet, in_args=in_args
        )

        wrapped_func = retry(
            wait=wait_random_exponential(min=1, max=8),
            stop=stop_after_attempt(self._num_par_run_retries + 1),
            before_sleep=retry_before_sleep_callback,
            retry_error_callback=retry_error_callback,
        )(self._run_single)

        tasks = [
            wrapped_func(
                in_args=inp,
                forgetful=True,
                run_id=self._generate_par_run_id(run_id, idx),
                ctx=ctx,
            )
            for idx, inp in enumerate(par_inputs)
        ]
        out_packets = await asyncio.gather(*tasks)

        return Packet(  # type: ignore[return]
            payloads=[
                (out_packet.payloads[0] if out_packet else None)
                for out_packet in out_packets
            ],
            sender=self.name,
        )

    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT_co]:
        if isinstance(in_args, Sequence):
            return await self._run_par(
                chat_inputs=chat_inputs,
                in_packet=in_packet,
                in_args=cast("Sequence[InT]", in_args),
                run_id=run_id,
                forgetful=forgetful,
                ctx=ctx,
            )
        return await self._run_single(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            forgetful=forgetful,
            run_id=run_id,
            ctx=ctx,
        )

    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        resolved_in_args = self._validate_and_resolve_single_input(
            chat_inputs=chat_inputs, in_packet=in_packet, in_args=in_args
        )

        _memory = self.memory.model_copy(deep=True) if forgetful else self.memory

        outputs: Sequence[OutT_co] = []
        async for output_event in self._process_stream(
            chat_inputs=chat_inputs,
            in_args=resolved_in_args,
            memory=_memory,
            run_id=self._generate_run_id(run_id),
            ctx=ctx,
        ):
            if isinstance(output_event, ProcOutputEvent):
                outputs.append(output_event.data)
            else:
                yield output_event

        val_outputs = self._validate_outputs(outputs)
        out_packet = Packet[OutT_co](payloads=val_outputs, sender=self.name)

        yield PacketEvent(data=out_packet, name=self.name)

    @final
    def as_tool(
        self, tool_name: str, tool_description: str
    ) -> BaseTool[InT, OutT_co, Any]:  # type: ignore[override]
        # TODO: stream tools
        processor_instance = self
        in_type = processor_instance.in_type
        out_type = processor_instance.out_type
        if not issubclass(in_type, BaseModel):
            raise TypeError(
                "Cannot create a tool from an agent with "
                f"non-BaseModel input type: {in_type}"
            )

        class ProcessorTool(BaseTool[in_type, out_type, Any]):
            name: str = tool_name
            description: str = tool_description

            async def run(
                self, inp: InT, ctx: RunContext[CtxT] | None = None
            ) -> OutT_co:
                result = await processor_instance.run(
                    in_args=in_type.model_validate(inp), forgetful=True, ctx=ctx
                )

                return result.payloads[0]

        return ProcessorTool()  # type: ignore[return-value]



from typing import Any

from openai import pydantic_function_tool
from pydantic import BaseModel

from ..typing.tool import BaseTool, NamedToolChoice, ToolChoice
from . import (
    OpenAIFunctionDefinition,
    OpenAINamedToolChoiceFunction,
    OpenAINamedToolChoiceParam,
    OpenAIToolChoiceOptionParam,
    OpenAIToolParam,
)


def to_api_tool(
    tool: BaseTool[BaseModel, Any, Any], strict: bool | None = None
) -> OpenAIToolParam:
    if strict:
        return pydantic_function_tool(
            model=tool.in_type, name=tool.name, description=tool.description
        )

    function = OpenAIFunctionDefinition(
        name=tool.name,
        description=tool.description,
        parameters=tool.in_type.model_json_schema(),
        strict=strict,
    )
    if strict is None:
        function.pop("strict")

    return OpenAIToolParam(type="function", function=function)


def to_api_tool_choice(tool_choice: ToolChoice) -> OpenAIToolChoiceOptionParam:
    if isinstance(tool_choice, NamedToolChoice):
        return OpenAINamedToolChoiceParam(
            type="function",
            function=OpenAINamedToolChoiceFunction(name=tool_choice.name),
        )
    return tool_choice



# pyright: reportUnusedImport=false

from openai._streaming import (
    AsyncStream as OpenAIAsyncStream,  # type: ignore[import] # noqa: PLC2701
)
from openai.types import CompletionUsage as OpenAICompletionUsage
from openai.types.chat.chat_completion import ChatCompletion as OpenAICompletion
from openai.types.chat.chat_completion import (
    ChoiceLogprobs as OpenAIChoiceLogprobs,
)
from openai.types.chat.chat_completion_assistant_message_param import (
    ChatCompletionAssistantMessageParam as OpenAIAssistantMessageParam,
)
from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk as OpenAICompletionChunk,
)
from openai.types.chat.chat_completion_chunk import (
    Choice as CompletionChunkChoice,
)
from openai.types.chat.chat_completion_chunk import (
    Choice as OpenAIChunkChoice,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDelta as CompletionChunkChoiceDelta,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDelta as OpenAIChunkChoiceDelta,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDeltaToolCall as OpenAIChunkChoiceDeltaToolCall,
)
from openai.types.chat.chat_completion_content_part_image_param import (
    ChatCompletionContentPartImageParam as OpenAIContentPartImageParam,
)
from openai.types.chat.chat_completion_content_part_image_param import (
    ImageURL as OpenAIImageURL,
)
from openai.types.chat.chat_completion_content_part_param import (
    ChatCompletionContentPartParam as OpenAIContentPartParam,
)
from openai.types.chat.chat_completion_content_part_text_param import (
    ChatCompletionContentPartTextParam as OpenAIContentPartTextParam,
)
from openai.types.chat.chat_completion_developer_message_param import (
    ChatCompletionDeveloperMessageParam as OpenAIDeveloperMessageParam,
)
from openai.types.chat.chat_completion_function_message_param import (
    ChatCompletionFunctionMessageParam as OpenAIFunctionMessageParam,
)
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage as OpenAIMessage,
)
from openai.types.chat.chat_completion_message_param import (
    ChatCompletionMessageParam as OpenAIMessageParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import (
    ChatCompletionMessageToolCallParam as OpenAIToolCallParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import (
    Function as OpenAIToolCallFunction,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    ChatCompletionNamedToolChoiceParam as OpenAINamedToolChoiceParam,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    Function as OpenAINamedToolChoiceFunction,
)
from openai.types.chat.chat_completion_prediction_content_param import (
    ChatCompletionPredictionContentParam as OpenAIPredictionContentParam,
)
from openai.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam as OpenAIStreamOptionsParam,
)
from openai.types.chat.chat_completion_system_message_param import (
    ChatCompletionSystemMessageParam as OpenAISystemMessageParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam as OpenAIToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_message_param import (
    ChatCompletionToolMessageParam as OpenAIToolMessageParam,
)
from openai.types.chat.chat_completion_tool_param import (
    ChatCompletionToolParam as OpenAIToolParam,
)
from openai.types.chat.chat_completion_user_message_param import (
    ChatCompletionUserMessageParam as OpenAIUserMessageParam,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChatCompletion as OpenAIParsedCompletion,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChatCompletionMessage as OpenAIParsedMessage,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChoice as OpenAIParsedChoice,
)
from openai.types.shared_params.function_definition import (
    FunctionDefinition as OpenAIFunctionDefinition,
)

# from openai.beta.types.chat.
from .openai_llm import OpenAILLM, OpenAILLMSettings

__all__ = ["OpenAILLM", "OpenAILLMSettings"]



import time
from enum import StrEnum
from typing import Any, Generic, Literal, TypeVar

from pydantic import BaseModel, Field

from ..packet import Packet
from .completion import Completion
from .completion_chunk import CompletionChunk
from .message import AssistantMessage, SystemMessage, ToolCall, ToolMessage, UserMessage


class EventSourceType(StrEnum):
    LLM = "llm"
    AGENT = "agent"
    USER = "user"
    TOOL = "tool"
    PROCESSOR = "processor"


class EventType(StrEnum):
    SYS_MSG = "system_message"
    USR_MSG = "user_message"
    TOOL_MSG = "tool_message"
    TOOL_CALL = "tool_call"
    GEN_MSG = "gen_message"
    COMP = "completion"
    COMP_CHUNK = "completion_chunk"
    PACKET = "packet"
    PROC_OUT = "processor_output"


_T = TypeVar("_T")


class Event(BaseModel, Generic[_T], frozen=True):
    type: EventType
    source: EventSourceType
    created: int = Field(default_factory=lambda: int(time.time()))
    name: str | None = None
    data: _T


class CompletionEvent(Event[Completion], frozen=True):
    type: Literal[EventType.COMP] = EventType.COMP
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class CompletionChunkEvent(Event[CompletionChunk], frozen=True):
    type: Literal[EventType.COMP_CHUNK] = EventType.COMP_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class GenMessageEvent(Event[AssistantMessage], frozen=True):
    type: Literal[EventType.GEN_MSG] = EventType.GEN_MSG
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class ToolCallEvent(Event[ToolCall], frozen=True):
    type: Literal[EventType.TOOL_CALL] = EventType.TOOL_CALL
    source: Literal[EventSourceType.AGENT] = EventSourceType.AGENT


class ToolMessageEvent(Event[ToolMessage], frozen=True):
    type: Literal[EventType.TOOL_MSG] = EventType.TOOL_MSG
    source: Literal[EventSourceType.TOOL] = EventSourceType.TOOL


class UserMessageEvent(Event[UserMessage], frozen=True):
    type: Literal[EventType.USR_MSG] = EventType.USR_MSG
    source: Literal[EventSourceType.USER] = EventSourceType.USER


class SystemMessageEvent(Event[SystemMessage], frozen=True):
    type: Literal[EventType.SYS_MSG] = EventType.SYS_MSG
    source: Literal[EventSourceType.AGENT] = EventSourceType.AGENT


class PacketEvent(Event[Packet[Any]], frozen=True):
    type: Literal[EventType.PACKET] = EventType.PACKET
    source: Literal[EventSourceType.PROCESSOR] = EventSourceType.PROCESSOR


class ProcOutputEvent(Event[Any], frozen=True):
    type: Literal[EventType.PROC_OUT] = EventType.PROC_OUT
    source: Literal[EventSourceType.PROCESSOR] = EventSourceType.PROCESSOR



import hashlib
import json
import logging
from collections.abc import Mapping, Sequence
from typing import Literal, TypeAlias

from termcolor._types import Color  # type: ignore[import]

from .typing.completion import Usage
from .typing.content import Content, ContentPartText
from .typing.message import AssistantMessage, Message, Role, ToolMessage

logger = logging.getLogger(__name__)


ColoringMode: TypeAlias = Literal["agent", "role"]

ROLE_TO_COLOR: Mapping[Role, Color] = {
    Role.SYSTEM: "magenta",
    Role.USER: "green",
    Role.ASSISTANT: "light_blue",
    Role.TOOL: "light_cyan",
}

AVAILABLE_COLORS: list[Color] = [
    "magenta",
    "green",
    "light_blue",
    "light_cyan",
    "yellow",
    "blue",
    "red",
]


class Printer:
    def __init__(
        self,
        color_by: ColoringMode = "role",
        msg_trunc_len: int = 20000,
        print_messages: bool = False,
    ) -> None:
        self.color_by = color_by
        self.msg_trunc_len = msg_trunc_len
        self.print_messages = print_messages

    @staticmethod
    def get_role_color(role: Role) -> Color:
        return ROLE_TO_COLOR[role]

    @staticmethod
    def get_agent_color(agent_name: str) -> Color:
        idx = int(
            hashlib.md5(agent_name.encode()).hexdigest(),  # noqa :S324
            16,
        ) % len(AVAILABLE_COLORS)

        return AVAILABLE_COLORS[idx]

    @staticmethod
    def content_to_str(content: Content | str, role: Role) -> str:
        if role == Role.USER and isinstance(content, Content):
            content_str_parts: list[str] = []
            for content_part in content.parts:
                if isinstance(content_part, ContentPartText):
                    content_str_parts.append(content_part.data.strip(" \n"))
                elif content_part.data.type == "url":
                    content_str_parts.append(str(content_part.data.url))
                elif content_part.data.type == "base64":
                    content_str_parts.append("<ENCODED_IMAGE>")
            return "\n".join(content_str_parts)

        assert isinstance(content, str)

        return content.strip(" \n")

    @staticmethod
    def truncate_content_str(content_str: str, trunc_len: int = 2000) -> str:
        if len(content_str) > trunc_len:
            return content_str[:trunc_len] + "[...]"

        return content_str

    def print_llm_message(
        self, message: Message, agent_name: str, run_id: str, usage: Usage | None = None
    ) -> None:
        if not self.print_messages:
            return

        if usage is not None and not isinstance(message, AssistantMessage):
            raise ValueError(
                "Usage information can only be printed for AssistantMessage"
            )

        role = message.role
        content_str = self.content_to_str(message.content or "", message.role)

        if self.color_by == "agent":
            color = self.get_agent_color(agent_name)
        elif self.color_by == "role":
            color = self.get_role_color(role)

        log_kwargs = {"extra": {"color": color}}  # type: ignore

        # Print message title

        out = f"\n[agent: {agent_name} | role: {role.value} | run: {run_id}]"

        if isinstance(message, ToolMessage):
            out += f"\n{message.name} | {message.tool_call_id}"

        # Print message content

        if content_str:
            try:
                content_str = json.dumps(json.loads(content_str), indent=2)
            except Exception:
                pass
            content_str_truncated = self.truncate_content_str(
                content_str, trunc_len=self.msg_trunc_len
            )
            out += f"\n{content_str_truncated}"

        logger.debug(out, **log_kwargs)  # type: ignore

        # Print tool calls

        if isinstance(message, AssistantMessage) and message.tool_calls is not None:
            for tool_call in message.tool_calls:
                if self.color_by == "agent":
                    tool_color = self.get_agent_color(agent_name=agent_name)
                elif self.color_by == "role":
                    tool_color = self.get_role_color(role=Role.TOOL)
                logger.debug(
                    f"\n<{agent_name}>[TOOL_CALL]\n{tool_call.tool_name} "
                    f"| {tool_call.id}\n{tool_call.tool_arguments}",
                    extra={"color": tool_color},  # type: ignore
                )

        # Print usage

        if usage is not None:
            usage_str = (
                f"I/O/(R)/(C) tokens: {usage.input_tokens}/{usage.output_tokens}"
            )
            if usage.reasoning_tokens is not None:
                usage_str += f"/{usage.reasoning_tokens}"
            if usage.cached_tokens is not None:
                usage_str += f"/{usage.cached_tokens}"
            logger.debug(
                f"\n------------------------------------\n{usage_str}",
                **log_kwargs,  # type: ignore
            )

    def print_llm_messages(
        self,
        messages: Sequence[Message],
        agent_name: str,
        run_id: str,
        usages: Sequence[Usage | None] | None = None,
    ) -> None:
        if not self.print_messages:
            return

        _usages: Sequence[Usage | None] = usages or [None] * len(messages)

        for _message, _usage in zip(messages, _usages, strict=False):
            self.print_llm_message(
                _message, usage=_usage, agent_name=agent_name, run_id=run_id
            )



import asyncio
import logging
from typing import Any, Generic, Protocol, TypeVar

from .packet import Packet
from .run_context import CtxT, RunContext
from .typing.io import ProcName

logger = logging.getLogger(__name__)


_PayloadT_contra = TypeVar("_PayloadT_contra", contravariant=True)


class PacketHandler(Protocol[_PayloadT_contra, CtxT]):
    async def __call__(
        self,
        packet: Packet[_PayloadT_contra],
        ctx: RunContext[CtxT] | None,
        **kwargs: Any,
    ) -> None: ...


class PacketPool(Generic[CtxT]):
    def __init__(self) -> None:
        self._queues: dict[ProcName, asyncio.Queue[Packet[Any]]] = {}
        self._packet_handlers: dict[ProcName, PacketHandler[Any, CtxT]] = {}
        self._tasks: dict[ProcName, asyncio.Task[None]] = {}

    async def post(self, packet: Packet[Any]) -> None:
        for recipient_id in packet.recipients:
            queue = self._queues.setdefault(recipient_id, asyncio.Queue())
            await queue.put(packet)

    def register_packet_handler(
        self,
        processor_name: ProcName,
        handler: PacketHandler[Any, CtxT],
        ctx: RunContext[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        self._packet_handlers[processor_name] = handler
        self._queues.setdefault(processor_name, asyncio.Queue())
        if processor_name not in self._tasks:
            self._tasks[processor_name] = asyncio.create_task(
                self._handle_packets(processor_name, ctx=ctx, **run_kwargs)
            )

    async def _handle_packets(
        self,
        processor_name: ProcName,
        ctx: RunContext[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        queue = self._queues[processor_name]
        while True:
            try:
                packet = await queue.get()
                handler = self._packet_handlers.get(processor_name)
                if handler is None:
                    break

                try:
                    await self._packet_handlers[processor_name](
                        packet, ctx=ctx, **run_kwargs
                    )
                except Exception:
                    logger.exception(f"Error handling packet for {processor_name}")

                queue.task_done()

            except Exception:
                logger.exception(
                    f"Unexpected error in processing loop for {processor_name}"
                )

    async def unregister_packet_handler(self, processor_name: ProcName) -> None:
        if task := self._tasks.get(processor_name):
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.info(f"{processor_name} exited")

        self._tasks.pop(processor_name, None)
        self._queues.pop(processor_name, None)
        self._packet_handlers.pop(processor_name, None)

    async def stop_all(self) -> None:
        for processor_name in list(self._tasks):
            await self.unregister_packet_handler(processor_name)



from collections import defaultdict
from collections.abc import Mapping
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from grasp_agents.typing.completion import Completion

from .printer import ColoringMode, Printer
from .typing.io import LLMPromptArgs, ProcName
from .usage_tracker import UsageTracker


class RunArgs(BaseModel):
    sys: LLMPromptArgs | None = None
    usr: LLMPromptArgs | None = None

    model_config = ConfigDict(extra="forbid")


CtxT = TypeVar("CtxT")


class RunContext(BaseModel, Generic[CtxT]):
    state: CtxT | None = None

    run_args: dict[ProcName, RunArgs] = Field(default_factory=dict)
    completions: Mapping[ProcName, list[Completion]] = Field(
        default_factory=lambda: defaultdict(list)
    )

    print_messages: bool = False
    color_messages_by: ColoringMode = "role"

    _usage_tracker: UsageTracker = PrivateAttr()
    _printer: Printer = PrivateAttr()

    def model_post_init(self, context: Any) -> None:  # noqa: ARG002
        self._usage_tracker = UsageTracker()
        self._printer = Printer(
            print_messages=self.print_messages,
            color_by=self.color_messages_by,
        )

    @property
    def usage_tracker(self) -> UsageTracker:
        return self._usage_tracker

    @property
    def printer(self) -> Printer:
        return self._printer

    model_config = ConfigDict(extra="forbid")



class InputValidationError(Exception):
    pass


class StringParsingError(Exception):
    pass


class CompletionError(Exception):
    pass


class CombineCompletionChunksError(Exception):
    pass


class ToolValidationError(Exception):
    pass


class OutputValidationError(Exception):
    pass


class WorkflowConstructionError(Exception):
    pass


class SystemPromptBuilderError(Exception):
    pass


class InputPromptBuilderError(Exception):
    pass



from collections.abc import Sequence
from itertools import pairwise
from typing import Any, ClassVar, Generic, cast, final

from ..errors import WorkflowConstructionError
from ..packet_pool import Packet, PacketPool
from ..processor import Processor
from ..run_context import CtxT, RunContext
from ..typing.io import InT, OutT_co, ProcName
from .workflow_processor import WorkflowProcessor


class SequentialWorkflow(
    WorkflowProcessor[InT, OutT_co, CtxT], Generic[InT, OutT_co, CtxT]
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[Processor[Any, Any, Any, CtxT]],
        packet_pool: PacketPool[CtxT] | None = None,
        recipients: list[ProcName] | None = None,
        num_par_run_retries: int = 0,
    ) -> None:
        super().__init__(
            subprocs=subprocs,
            start_proc=subprocs[0],
            end_proc=subprocs[-1],
            name=name,
            packet_pool=packet_pool,
            recipients=recipients,
            num_par_run_retries=num_par_run_retries,
        )

        for prev_proc, proc in pairwise(subprocs):
            if prev_proc.out_type != proc.in_type:
                raise WorkflowConstructionError(
                    f"Output type {prev_proc.out_type} of subprocessor {prev_proc.name}"
                    f" does not match input type {proc.in_type} of subprocessor"
                    f" {proc.name}"
                )

    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT_co]:
        packet = in_packet
        for subproc in self.subprocs:
            packet = await subproc.run(
                chat_inputs=chat_inputs,
                in_packet=packet,
                in_args=in_args,
                forgetful=forgetful,
                run_id=self._generate_subproc_run_id(run_id, subproc),
                ctx=ctx,
            )
            chat_inputs = None
            in_args = None

        return cast("Packet[OutT_co]", packet)



from collections.abc import AsyncIterator, Sequence
from pathlib import Path
from typing import Any, ClassVar, Generic, Protocol, TypeVar

from pydantic import BaseModel

from .comm_processor import CommProcessor
from .llm import LLM, LLMSettings
from .llm_agent_memory import LLMAgentMemory, MakeMemoryHandler
from .llm_policy_executor import (
    ExitToolCallLoopHandler,
    LLMPolicyExecutor,
    ManageMemoryHandler,
)
from .packet_pool import PacketPool
from .prompt_builder import (
    MakeInputContentHandler,
    MakeSystemPromptHandler,
    PromptBuilder,
)
from .run_context import CtxT, RunContext
from .typing.content import Content, ImageData
from .typing.converters import Converters
from .typing.events import Event, ProcOutputEvent, SystemMessageEvent, UserMessageEvent
from .typing.io import InT, LLMPrompt, LLMPromptArgs, OutT_co, ProcName
from .typing.message import Message, Messages, SystemMessage, UserMessage
from .typing.tool import BaseTool
from .utils import get_prompt, validate_obj_from_json_or_py_string

_InT_contra = TypeVar("_InT_contra", contravariant=True)
_OutT_co = TypeVar("_OutT_co", covariant=True)


class ParseOutputHandler(Protocol[_InT_contra, _OutT_co, CtxT]):
    def __call__(
        self,
        conversation: Messages,
        *,
        in_args: _InT_contra | None,
        ctx: RunContext[CtxT] | None,
    ) -> _OutT_co: ...


class LLMAgent(
    CommProcessor[InT, OutT_co, LLMAgentMemory, CtxT],
    Generic[InT, OutT_co, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        *,
        # LLM
        llm: LLM[LLMSettings, Converters],
        # Tools
        tools: list[BaseTool[Any, Any, CtxT]] | None = None,
        # Input prompt template (combines user and received arguments)
        in_prompt: LLMPrompt | None = None,
        in_prompt_path: str | Path | None = None,
        # System prompt template
        sys_prompt: LLMPrompt | None = None,
        sys_prompt_path: str | Path | None = None,
        # System args (static args provided via RunContext)
        sys_args_schema: type[LLMPromptArgs] | None = None,
        # User args (static args provided via RunContext)
        usr_args_schema: type[LLMPromptArgs] | None = None,
        # Agent loop settings
        max_turns: int = 100,
        react_mode: bool = False,
        final_answer_as_tool_call: bool = False,
        # Agent memory management
        reset_memory_on_run: bool = False,
        # Retries
        num_par_run_retries: int = 0,
        # Multi-agent routing
        packet_pool: PacketPool[CtxT] | None = None,
        recipients: list[ProcName] | None = None,
    ) -> None:
        super().__init__(
            name=name,
            packet_pool=packet_pool,
            recipients=recipients,
            num_par_run_retries=num_par_run_retries,
        )

        # Agent memory

        self._memory: LLMAgentMemory = LLMAgentMemory()
        self._reset_memory_on_run = reset_memory_on_run

        # LLM policy executor

        self._used_default_llm_response_format: bool = False
        if llm.response_format is None and tools is None:
            llm.response_format = self.out_type
            self._used_default_llm_response_format = True

        self._policy_executor: LLMPolicyExecutor[OutT_co, CtxT] = LLMPolicyExecutor[
            self.out_type, CtxT
        ](
            agent_name=self.name,
            llm=llm,
            tools=tools,
            max_turns=max_turns,
            react_mode=react_mode,
            final_answer_as_tool_call=final_answer_as_tool_call,
        )

        # Prompt builder

        sys_prompt = get_prompt(prompt_text=sys_prompt, prompt_path=sys_prompt_path)
        in_prompt = get_prompt(prompt_text=in_prompt, prompt_path=in_prompt_path)
        self._prompt_builder: PromptBuilder[InT, CtxT] = PromptBuilder[
            self.in_type, CtxT
        ](
            agent_name=self._name,
            sys_prompt_template=sys_prompt,
            in_prompt_template=in_prompt,
            sys_args_schema=sys_args_schema,
            usr_args_schema=usr_args_schema,
        )

        # self.no_tqdm = getattr(llm, "no_tqdm", False)

        self._make_memory_impl: MakeMemoryHandler | None = None
        self._parse_output_impl: ParseOutputHandler[InT, OutT_co, CtxT] | None = None
        self._register_overridden_handlers()

    @property
    def llm(self) -> LLM[LLMSettings, Converters]:
        return self._policy_executor.llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._policy_executor.tools

    @property
    def max_turns(self) -> int:
        return self._policy_executor.max_turns

    @property
    def sys_args_schema(self) -> type[LLMPromptArgs] | None:
        return self._prompt_builder.sys_args_schema

    @property
    def usr_args_schema(self) -> type[LLMPromptArgs] | None:
        return self._prompt_builder.usr_args_schema

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.sys_prompt_template

    @property
    def in_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.in_prompt_template

    def _memorize_inputs(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT] | None = None,
    ) -> tuple[SystemMessage | None, UserMessage | None, LLMAgentMemory]:
        # 1. Get run arguments
        sys_args: LLMPromptArgs | None = None
        usr_args: LLMPromptArgs | None = None
        if ctx is not None:
            run_args = ctx.run_args.get(self.name)
            if run_args is not None:
                sys_args = run_args.sys
                usr_args = run_args.usr

        # 2. Make system prompt (can be None)

        formatted_sys_prompt = self._prompt_builder.make_system_prompt(
            sys_args=sys_args, ctx=ctx
        )

        # 3. Set agent memory

        system_message: SystemMessage | None = None
        if self._reset_memory_on_run or memory.is_empty:
            memory.reset(formatted_sys_prompt)
            if formatted_sys_prompt is not None:
                system_message = memory.message_history[0]  # type: ignore[union-attr]
        elif self._make_memory_impl:
            memory = self._make_memory_impl(
                prev_memory=memory,
                in_args=in_args,
                sys_prompt=formatted_sys_prompt,
                ctx=ctx,
            )

        # 3. Make and add user messages

        user_message = self._prompt_builder.make_user_message(
            chat_inputs=chat_inputs, in_args=in_args, usr_args=usr_args, ctx=ctx
        )
        if user_message:
            memory.update([user_message])

        return system_message, user_message, memory

    def _parse_output(
        self,
        conversation: Messages,
        *,
        in_args: InT | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> OutT_co:
        if self._parse_output_impl:
            return self._parse_output_impl(
                conversation=conversation, in_args=in_args, ctx=ctx
            )

        return validate_obj_from_json_or_py_string(
            str(conversation[-1].content or ""),
            adapter=self._out_type_adapter,
            from_substring=False,
        )

    async def _process(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> Sequence[OutT_co]:
        system_message, user_message, memory = self._memorize_inputs(
            chat_inputs=chat_inputs, in_args=in_args, memory=memory, ctx=ctx
        )
        if system_message:
            self._print_messages([system_message], run_id=run_id, ctx=ctx)
        if user_message:
            self._print_messages([user_message], run_id=run_id, ctx=ctx)

        await self._policy_executor.execute(memory, run_id=run_id, ctx=ctx)

        return [
            self._parse_output(
                conversation=memory.message_history, in_args=in_args, ctx=ctx
            )
        ]

    async def _process_stream(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        system_message, user_message, memory = self._memorize_inputs(
            chat_inputs=chat_inputs, in_args=in_args, memory=memory, ctx=ctx
        )
        if system_message:
            yield SystemMessageEvent(data=system_message)
        if user_message:
            yield UserMessageEvent(data=user_message)

        # 4. Run tool call loop (new messages are added to the message
        #    history inside the loop)
        async for event in self._policy_executor.execute_stream(
            memory, run_id=run_id, ctx=ctx
        ):
            yield event

        output = self._parse_output(
            conversation=memory.message_history, in_args=in_args, ctx=ctx
        )
        yield ProcOutputEvent(data=output, name=self.name)

    def _print_messages(
        self,
        messages: Sequence[Message],
        run_id: str,
        ctx: RunContext[CtxT] | None = None,
    ) -> None:
        if ctx:
            ctx.printer.print_llm_messages(
                messages, agent_name=self.name, run_id=run_id
            )

    # -- Decorators for custom implementations --

    def make_system_prompt(
        self, func: MakeSystemPromptHandler[CtxT]
    ) -> MakeSystemPromptHandler[CtxT]:
        self._prompt_builder.make_system_prompt_impl = func

        return func

    def make_input_content(
        self, func: MakeInputContentHandler[InT, CtxT]
    ) -> MakeInputContentHandler[InT, CtxT]:
        self._prompt_builder.make_input_content_impl = func

        return func

    def parse_output(
        self, func: ParseOutputHandler[InT, OutT_co, CtxT]
    ) -> ParseOutputHandler[InT, OutT_co, CtxT]:
        if self._used_default_llm_response_format:
            self._policy_executor.llm.response_format = None
        self._parse_output_impl = func

        return func

    def make_memory(self, func: MakeMemoryHandler) -> MakeMemoryHandler:
        self._make_memory_impl = func

        return func

    def manage_memory(
        self, func: ManageMemoryHandler[CtxT]
    ) -> ManageMemoryHandler[CtxT]:
        self._policy_executor.manage_memory_impl = func

        return func

    def exit_tool_call_loop(
        self, func: ExitToolCallLoopHandler[CtxT]
    ) -> ExitToolCallLoopHandler[CtxT]:
        self._policy_executor.exit_tool_call_loop_impl = func

        return func

    # -- Override these methods in subclasses if needed --

    def _register_overridden_handlers(self) -> None:
        cur_cls = type(self)
        base_cls = LLMAgent[Any, Any, Any]

        if cur_cls._make_system_prompt is not base_cls._make_system_prompt:  # noqa: SLF001
            self._prompt_builder.make_system_prompt_impl = self._make_system_prompt

        if cur_cls._make_input_content is not base_cls._make_input_content:  # noqa: SLF001
            self._prompt_builder.make_input_content_impl = self._make_input_content

        if cur_cls._make_memory is not base_cls._make_memory:  # noqa: SLF001
            self._make_memory_impl = self._make_memory

        if cur_cls._manage_memory is not base_cls._manage_memory:  # noqa: SLF001
            self._policy_executor.manage_memory_impl = self._manage_memory

        if (
            cur_cls._exit_tool_call_loop is not base_cls._exit_tool_call_loop  # noqa: SLF001
        ):
            self._policy_executor.exit_tool_call_loop_impl = self._exit_tool_call_loop

        if (
            cur_cls._parse_output is not base_cls._parse_output  # noqa: SLF001
            and self._used_default_llm_response_format
        ):
            self._policy_executor.llm.response_format = None

    def _make_system_prompt(
        self, sys_args: LLMPromptArgs | None, *, ctx: RunContext[CtxT] | None = None
    ) -> str:
        raise NotImplementedError(
            "LLMAgent._format_sys_args must be overridden by a subclass "
            "if it's intended to be used as the system arguments formatter."
        )

    def _make_input_content(
        self,
        *,
        in_args: InT | None = None,
        usr_args: LLMPromptArgs | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Content:
        raise NotImplementedError(
            "LLMAgent._format_in_args must be overridden by a subclass"
        )

    def _make_memory(
        self,
        prev_memory: LLMAgentMemory,
        in_args: Sequence[InT] | None = None,
        sys_prompt: LLMPrompt | None = None,
        ctx: RunContext[Any] | None = None,
    ) -> LLMAgentMemory:
        raise NotImplementedError(
            "LLMAgent._make_memory must be overridden by a subclass"
        )

    def _exit_tool_call_loop(
        self,
        conversation: Messages,
        *,
        ctx: RunContext[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        raise NotImplementedError(
            "LLMAgent._exit_tool_call_loop must be overridden by a subclass"
        )

    def _manage_memory(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT] | None = None,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError(
            "LLMAgent._manage_memory must be overridden by a subclass"
        )



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, ClassVar, Generic

from ..comm_processor import CommProcessor
from ..errors import WorkflowConstructionError
from ..packet import Packet
from ..packet_pool import PacketPool
from ..processor import Processor
from ..run_context import CtxT, RunContext
from ..typing.io import InT, OutT_co, ProcName


class WorkflowProcessor(
    CommProcessor[InT, OutT_co, Any, CtxT],
    ABC,
    Generic[InT, OutT_co, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[Processor[Any, Any, Any, CtxT]],
        start_proc: Processor[InT, Any, Any, CtxT],
        end_proc: Processor[Any, OutT_co, Any, CtxT],
        packet_pool: PacketPool[CtxT] | None = None,
        recipients: list[ProcName] | None = None,
        num_par_run_retries: int = 0,
    ) -> None:
        super().__init__(
            name=name,
            packet_pool=packet_pool,
            recipients=recipients,
            num_par_run_retries=num_par_run_retries,
        )

        if len(subprocs) < 2:
            raise WorkflowConstructionError("At least two subprocessors are required")
        if start_proc not in subprocs:
            raise WorkflowConstructionError(
                "Start subprocessor must be in the subprocessors list"
            )
        if end_proc not in subprocs:
            raise WorkflowConstructionError(
                "End subprocessor must be in the subprocessors list"
            )

        if start_proc.in_type != self.in_type:
            raise WorkflowConstructionError(
                f"Start subprocessor's input type {start_proc.in_type} does not "
                f"match workflow's input type {self._in_type}"
            )
        if end_proc.out_type != self.out_type:
            raise WorkflowConstructionError(
                f"End subprocessor's output type {end_proc.out_type} does not "
                f"match workflow's output type {self._out_type}"
            )

        self._subprocs = subprocs
        self._start_proc = start_proc
        self._end_proc = end_proc

    @property
    def subprocs(self) -> Sequence[Processor[Any, Any, Any, CtxT]]:
        return self._subprocs

    @property
    def start_proc(self) -> Processor[InT, Any, Any, CtxT]:
        return self._start_proc

    @property
    def end_proc(self) -> Processor[Any, OutT_co, Any, CtxT]:
        return self._end_proc

    def _generate_subproc_run_id(
        self, run_id: str | None, subproc: Processor[Any, Any, Any, CtxT]
    ) -> str | None:
        return f"{self._generate_run_id(run_id)}/{subproc.name}"

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        ctx: RunContext[CtxT] | None = None,
        forgetful: bool = False,
        run_id: str | None = None,
    ) -> Packet[OutT_co]:
        pass
